// suspend.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"

_error _raise_error(HRESULT hr, LPTSTR pgzApiDesc) 
{
	_error oError;
	oError.hr = hr;
	oError.pgzApiDesc = pgzApiDesc;
	return oError;
}


HRESULT HeapFreeEx(LPVOID lpMem) 
{
	try
	{
		HANDLE hHeap = GetProcessHeap();
		if (INVALID_HANDLE_VALUE == hHeap) throw _raise_error(E_FAIL, _T("GetProcessHeap()"));
		
		#if defined(_DEBUG)
			// generates a brTest > 0 but a dwLastError = 0 likely because error was
		    // caused by passing a bad argument that was non-detrimental
			BOOL brTest = HeapFree(hHeap, 0, 0);  
			if (brTest) brTest = FALSE;  // to remove Level 4 warning
			DWORD dwLastErrorTest = GetLastError();
			if (dwLastErrorTest) dwLastErrorTest = FALSE;  // to remove Level 4 warning
			
			// throws an exception due to a bad argument hHeap
			// brTest = HeapFree(0, 0, lpMem);
			// dwLastErrorTest = GetLastError();
		#endif

		BOOL br = HeapFree(hHeap, 0, lpMem);
		// if (!br) throw _raise_error(E_FAIL, _T("HeapFree()"));
		// since it seems the opposite seems true on failure, i.e. br > 0, we will use
		// the following if (br) or GetLastError tests instead to see if call was successful
		if (br) throw _raise_error(E_FAIL, _T("HeapFree()"));
		// DWORD dwLastError = GetLastError();
		// if (dwLastError) throw (E_FAIL);
		
		throw _raise_error(S_OK, _T("The operation completed successfully"));
	}

	catch (_error oError) {
		#if defined(_DEBUG)
			_bstr_t sBuffer = oError.pgzApiDesc;
			HRESULT hr = oError.hr;
		#endif
		return oError.hr;
	}
}


HRESULT GetWinStaType(DWORD *pdwWinStaType) 
{
	HWINSTA hPrcWinSta = NULL;

	try {
		if (!pdwWinStaType) throw _raise_error(E_INVALIDARG, _T("InvalidArgument"));
		
		hPrcWinSta = GetProcessWindowStation();
		if (INVALID_HANDLE_VALUE == hPrcWinSta) throw _raise_error(HRESULT_FROM_WIN32(GetLastError()), _T("GetWindowStation"));
		*pdwWinStaType = WINSTA_UNKNOWN;

		_TCHAR gzBuffer[MAX_PATH] = _T("\0");
		#if defined(_DEBUG) && defined(_UNICODE)
			INT ir = MAX_PATH;
			ir = sizeof(gzBuffer);
			ir = _tcslen(gzBuffer);
		#endif
		DWORD cbReqd = 0;

		#if defined(_DEBUG)
			// option 1 - doesn't differentiate between the winsta0 gui console session and a winsta0 gui
			// terminal session but it does differentiate between these types and a telnet cli console session
			USEROBJECTFLAGS oUsrObjFlgs;	
			BOOL br = GetUserObjectInformation(hPrcWinSta,	  // handle to object
											   UOI_FLAGS,	  // type of information to retrieve
											   &oUsrObjFlgs,  // information buffer
											   sizeof(USEROBJECTFLAGS),  // size of the buffer
											   &cbReqd);	  // receives required buffer size
			if (!br || cbReqd > sizeof(USEROBJECTFLAGS)) throw _raise_error(HRESULT_FROM_WIN32(GetLastError()), _T("GetUserObjectInformation()"));
			WSF_VISIBLE && oUsrObjFlgs.dwFlags ? *pdwWinStaType = WINSTA_CONS_OR_TRMN : *pdwWinStaType = WINSTA_TELNET;

			// option 2 - doesn't differentiate between the winsta0 gui console session and a winsta0 gui
			// terminal session but it does differentiate between these types and a telnet cli console session
			*pdwWinStaType = WINSTA_UNKNOWN;
			memset(gzBuffer, 0, sizeof(gzBuffer));
			cbReqd = 0;
			br = GetUserObjectInformation(hPrcWinSta,	// handle to object
											   UOI_NAME,	// type of information to retrieve
											   &gzBuffer,	// information buffer
											   MAX_PATH,	// size of the buffer
											   &cbReqd);	// receives required buffer size
			if (!br || cbReqd > MAX_PATH) throw _raise_error(HRESULT_FROM_WIN32(GetLastError()), _T("GetUserObjectInformation()"));
			!_tcsicmp(_T("WinSta0"), gzBuffer) ? *pdwWinStaType = WINSTA_CONS_OR_TRMN : *pdwWinStaType = WINSTA_TELNET;

			*pdwWinStaType = WINSTA_UNKNOWN;
			memset(gzBuffer, 0, sizeof(gzBuffer));
			cbReqd = 0;
		#endif
			
		// option 3 - differentiates between the winsta0 gui console session, a winsta0 gui terminal
		// session and a telnet cli console session
		cbReqd = GetEnvironmentVariable(_T("SESSIONNAME"), gzBuffer, sizeof(gzBuffer));
		if (cbReqd > MAX_PATH) throw _raise_error(HRESULT_FROM_WIN32(GetLastError()), _T("GetEnvironmentVariable()"));
		if (!cbReqd)
			*pdwWinStaType = WINSTA_TELNET;  
			// assume above is the case for cli utilities but could be otherwise for non-cli utility cases 
		if (WINSTA_UNKNOWN == *pdwWinStaType && !_tcsicmp(_T("Console"), gzBuffer))
			*pdwWinStaType = WINSTA_CONSOLE;
		if (WINSTA_UNKNOWN == *pdwWinStaType && !_tcsncicmp(_T("RDP-Tcp#"), gzBuffer, _tcslen(_T("RDP-Tcp#"))))
			*pdwWinStaType = WINSTA_TERMINAL;		
		
		throw _raise_error(S_OK, _T("The operation completed successfully"));
	}

	catch (_error oError) {
		#if defined(_DEBUG)
			_bstr_t sBuffer = oError.pgzApiDesc;
			HRESULT hr = oError.hr;
		#endif

		if (INVALID_HANDLE_VALUE != hPrcWinSta) BOOL br = CloseHandle(hPrcWinSta);  // not documented as needed for this handle type

		return oError.hr;
	}
}


HRESULT ShowHelp(INT iArgC, LPTSTR *agzArgV, DWORD dwWinStaType) 
{
	HWND hDsktopWnd = NULL;

	try {
		_TCHAR gzBuffer[MAX_PATH] = _T("\0");

		_tcscpy(gzBuffer, PROGRAM_VERSION);
		_tcscat(gzBuffer, _T("\n"));
		_tprintf(gzBuffer);
		
		_tcscpy(gzBuffer, _T("\ndescription\n"));
		_tprintf(gzBuffer);
		
		_tcscpy(gzBuffer, _T("  This utility requests a graceful suspend of an advanced configuration and\n"));
		_tprintf(gzBuffer);
		
		_tcscpy(gzBuffer, _T("  power interface (acpi) / os directed power management (ospm) enabled host.\n"));
		_tprintf(gzBuffer);
		
		_tcscpy(gzBuffer, _T("  This can be useful on hosts where terminal services has removed the suspend\n"));
		_tprintf(gzBuffer);
		
		_tcscpy(gzBuffer, _T("  ui options and your hardware does not provide a physical button alternative\n"));
		_tprintf(gzBuffer);
	
		_tcscpy(gzBuffer, _T("  to trigger this action in the absence of these ui options.\n"));
		_tprintf(gzBuffer);
	
		_tcscpy(gzBuffer, _T("\nusage = "));

		// _tcscat(gzBuffer, PROGRAM_NAME);
		// or
		// [optionally] acquire program name programmatically from agzArgV[0] string
		// note - depending on how application was called agzArgV[0] can contain
		// just <progname>, <partialPath>\<progname> or <fullpath>\<progname>.exe
		// and these values are different when stepping through a debug build 
		// versus a release build
		LPTSTR pgzEndOfPath = NULL, pgzDotExe = NULL;
		#if defined(_DEBUG)
			_tprintf(_T("\ndebuginfo agzArgV[0] = %s\n"), agzArgV[0]);
			_tprintf(_T("debuginfo before _tcsrchr( , '\\'|':') pgzEndOfPath = %#x, string value = %s\n"), pgzEndOfPath, pgzEndOfPath);
			_tprintf(_T("debuginfo before _tcsrchr( ,'.') pgzDotExt = %#x, string value = %s\n"), pgzDotExe, pgzDotExe);
		#endif
		pgzEndOfPath = _tcsrchr(agzArgV[0], _T('\\'));  // i.e. <path>\<progname>
		if (!pgzEndOfPath) pgzEndOfPath = _tcsrchr(agzArgV[0], _T(':'));  // i.e. <drive>:<progname>
		#if defined(_DEBUG)
			_tprintf(_T("debuginfo after pgzEndOfPath = %#x, string value = %s\n"), pgzEndOfPath, pgzEndOfPath);
		#endif
		if (pgzEndOfPath) {
			pgzDotExe = _tcsrchr(pgzEndOfPath, _T('.'));
			#if defined(_DEBUG)
				_tprintf(_T("debuginfo after pgzDotExt = %#x, string value = %s\n"), pgzDotExe, pgzDotExe);
			#endif
			if (pgzDotExe)
				_tcsncat(gzBuffer, pgzEndOfPath + 1, pgzDotExe - pgzEndOfPath - 1);
			else
				_tcscat(gzBuffer, pgzEndOfPath + 1);
		}
		else {
			pgzDotExe = _tcsrchr(agzArgV[0], _T('.'));
			#if defined(_DEBUG)
				_tprintf(_T("debuginfo after pgzDotExt = %#x, string value = %s\n"), pgzDotExe, pgzDotExe);
			#endif
			if (pgzDotExe)
				_tcsncat(gzBuffer, agzArgV[0], pgzDotExe - agzArgV[0]);
			else
				_tcscat(gzBuffer, agzArgV[0]);
		}

		_tcscat(gzBuffer, _T(" /y\n"));
		_tprintf(gzBuffer);

		_tcscpy(gzBuffer, _T("\nwhere /y = execute a graceful suspend of host\n"));
		_tprintf(gzBuffer);
	
		// _TCHAR gzBuffer[MAX_PATH] = _T("\0");
		memset(gzBuffer, 0, sizeof(gzBuffer));
		LPTSTR pgzr = _tcscpy(gzBuffer, _T("\ncurrently executing from a winsta"));
		if (!pgzr) throw _raise_error(HRESULT_FROM_WIN32(GetLastError()), _T("_tcscpy()"));
		switch (dwWinStaType) {
		case WINSTA_CONSOLE:
			_tcscat(gzBuffer, _T("0 gui console session\n"));
			break;
		case WINSTA_TERMINAL:
			_tcscat(gzBuffer, _T("0 gui terminal session\n"));
			break;
		case WINSTA_CONS_OR_TRMN:
			_tcscat(gzBuffer, _T("0 gui console or terminal session    \n"));
			break;
		case WINSTA_TELNET:
			_tcscat(gzBuffer, _T(" non-gui telnet session\n"));
			break;
		default:  // case WINSTA_UNKNOWN:
			_tcscat(gzBuffer, _T(" non-gui unknown session\n"));
			break;
		}
		_tprintf(gzBuffer);
		
		// [optionally] throw up messagebox if this is winsta0 gui console session
		#if defined(_FOOBAR) // || defined(_DEBUG) || defined(NDEBUG)
			if (WINSTA_CONSOLE == dwWinStaType) {
				hDsktopWnd = GetDesktopWindow();
				if (INVALID_HANDLE_VALUE == hDsktopWnd) throw _raise_error(HRESULT_FROM_WIN32(GetLastError()), _T("GetDesktopWindow"));
				// br = SetForegroundWindow(hDsktopWnd);
				// if (!br) throw _raise_error(HRESULT_FROM_WIN32(GetLastError()), _T("SetForgroundWindow"));
				MessageBox(hDsktopWnd, gzBuffer, _T("Suspend"), MB_SETFOREGROUND);
				// if (INVALID_HANDLE_VALUE != hDsktopWnd) BOOL br = CloseHandle(hDsktopWnd);  // not documented as needed for this handle type
				if (INVALID_HANDLE_VALUE != hDsktopWnd) hDsktopWnd = 0;  // CloseHandle() generates and exception
			}
		#endif

		throw _raise_error(S_OK, _T("The operation completed successfully"));
	}

	catch (_error oError) {
		#if defined(_DEBUG)
			_bstr_t sBuffer = oError.pgzApiDesc;
			HRESULT hr = oError.hr;
		#endif
		return oError.hr;
	}
}


HRESULT EnablePrivilege(HANDLE hTok, LPCTSTR pgzPrv, BOOL bEnable, TOKEN_PRIVILEGES &oTokPrvOld)
// EnablePrivilege(HANDLE hTok, LPCTSTR pgzPrv, BOOL bEnable, TOKEN_PRIVILEGES *poTokPrvOld)  // perhaps should be this
{
	try {
		// fill out the request form 
		TOKEN_PRIVILEGES oTokPrv;
		oTokPrv.PrivilegeCount = 1;
		oTokPrv.Privileges[0].Attributes = bEnable ? SE_PRIVILEGE_ENABLED : 0;
		
		BOOL br = LookupPrivilegeValue(NULL, pgzPrv, &oTokPrv.Privileges[0].Luid);
		if (!br) throw _raise_error(HRESULT_FROM_WIN32(GetLastError()), _T("LookupPrivilegeValue"));

		// hTok must have been opened with these permissions
		//   TOKEN_QUERY - to get the old privileges setting
		//   TOKEN_ADJUST_PRIVILEGES - to adjust the privileges
		DWORD cbTokPrvOld = sizeof(oTokPrvOld);
		br = AdjustTokenPrivileges(hTok, FALSE, &oTokPrv, cbTokPrvOld, &oTokPrvOld, &cbTokPrvOld);
		if (!br) throw _raise_error(HRESULT_FROM_WIN32(GetLastError()), _T("AdjustTokenPrivileges"));

		// note - ensure you check return value for ERROR_NOT_ALL_ASSIGNED
		throw _raise_error(HRESULT_FROM_WIN32(GetLastError()), _T("The operation completed successfully"));
	}

	catch (_error oError) {
		#if defined(_DEBUG)
			_bstr_t sBuffer = oError.pgzApiDesc;
			HRESULT hr = oError.hr;
		#endif
		return oError.hr;
	}
}


HRESULT RestorePrivilege(HANDLE hTok, TOKEN_PRIVILEGES &oTokPrvOld)
// RestorePrivilege(HANDLE hTok, TOKEN_PRIVILEGES *poTokPrvOld)  // perhaps should be this
{
	try {
		BOOL br = AdjustTokenPrivileges(hTok, FALSE, &oTokPrvOld, 0, NULL, NULL);
		if (!br) _raise_error(HRESULT_FROM_WIN32(GetLastError()), _T("AdjustTokenPrivileges"));

		// note - ensure you check return value for ERROR_NOT_ALL_ASSIGNED
		throw _raise_error(HRESULT_FROM_WIN32(GetLastError()), _T("The operation completed successfully"));
	}

	catch (_error oError) {
		#if defined(_DEBUG)
			_bstr_t sBuffer = oError.pgzApiDesc;
			HRESULT hr = oError.hr;
		#endif
		return oError.hr;
	}
}


int main(int argc, char* argv[])
{
	HANDLE hThread = NULL, hThdToken = NULL, hProcess = NULL, hPrcToken = NULL;
	HWND hDsktopWnd = NULL;
	LPTSTR *agzArgV = NULL;  // defined outsided of try frame so we can access/free in catch frame
	DWORD dwWinStaType = WINSTA_UNKNOWN;

	try {
		// convert command line arguments ansi string array to a generic string array
		#if defined(_UNICODE)
			INT iArgC = 0;
			agzArgV = (LPWSTR *)CommandLineToArgvW(GetCommandLine(), &iArgC);
		#else  // (_MBCS)
			INT iArgC = argc;	
			agzArgV = argv;
		#endif
		if (!iArgC || !agzArgV) throw _raise_error(E_INVALIDARG, _T("generic text agzArvV assignment or creation"));

		DWORD dwCliSwitches = CLISWITCH_NONE;
		for (INT iIndex = 1; iIndex < iArgC; iIndex++) 
		{
			BOOL bFoundValidSwitch = FALSE;
			if (!_tcsnicmp(_T("/h"), agzArgV[iIndex], 2) ||
			    !_tcsnicmp(_T("h"), agzArgV[iIndex], 2)  ||
			    !_tcsnicmp(_T("/?"), agzArgV[iIndex], 2) ||
				!_tcsnicmp(_T("?"), agzArgV[iIndex], 2)) {
				if (CLISWITCH_NONE & dwCliSwitches)  // if set then unset CLISWITCH_NONE
					dwCliSwitches ^= CLISWITCH_NONE;
				dwCliSwitches |= CLISWITCH_HELP;
				bFoundValidSwitch = TRUE;
			}

			if (!_tcsnicmp(_T("/y"), agzArgV[iIndex], 2)) {
				if (CLISWITCH_NONE & dwCliSwitches)  // if set then unset CLISWITCH_NONE
					dwCliSwitches ^= CLISWITCH_NONE;
				dwCliSwitches |= CLISWITCH_SUSPEND;
				bFoundValidSwitch = TRUE;
			}

			if (!bFoundValidSwitch) {  // make note of invalid switches
				if (CLISWITCH_NONE & dwCliSwitches)  // if set then unset CLISWITCH_NONE
					dwCliSwitches ^= CLISWITCH_NONE;
				dwCliSwitches |= CLISWITCH_INVALID;
			}
		}
		
		// get process window station and so we know if application was run from a winsta0
		// gui console session, winsta0 gui terminal session or winsta non-gui telnet session
		HRESULT hr = GetWinStaType(&dwWinStaType);
		if (FAILED(hr)) throw _raise_error(hr, _T("GetWinStaType()"));

		if ((CLISWITCH_NONE & dwCliSwitches) ||
			(CLISWITCH_HELP & dwCliSwitches) ||
			(CLISWITCH_INVALID & dwCliSwitches)) {
			hr = ShowHelp(iArgC, agzArgV, dwWinStaType);
			if (FAILED(hr)) throw _raise_error(hr, _T("ShowUsage()"));
			throw _raise_error(S_OK, _T("The operation completed successfully"));
		}

		// proceed with suspend action
		_TCHAR gzBuffer[MAX_PATH] = _T("\0");
		LPTSTR pgzr = _tcscpy(gzBuffer, _T("\nexecuting a graceful system suspend from a winsta"));
		if (!pgzr) throw _raise_error(HRESULT_FROM_WIN32(GetLastError()), _T("_tcscpy()"));
		switch (dwWinStaType) {
		case WINSTA_CONSOLE:
			_tcscat(gzBuffer, _T("0 gui console session\n"));
			break;
		case WINSTA_TERMINAL:
			_tcscat(gzBuffer, _T("0 gui terminal session\n"));
			break;
		case WINSTA_CONS_OR_TRMN:
			_tcscat(gzBuffer, _T("0 gui console or terminal session    \n"));
			break;
		case WINSTA_TELNET:
			_tcscat(gzBuffer, _T(" non-gui telnet session\n"));
			break;
		default:  // case WINSTA_UNKNOWN:
			_tcscat(gzBuffer, _T(" non-gui unknown session\n"));
			break;
		}
		_tprintf(gzBuffer);
		
		/*
		// DWORD dwThreadId = GetCurrentThreadId();
		// if (!dwThreadId) throw _raise_error(HRESULT_FROM_WIN32(GetLastError()), _T("GetCurrentThreadId"));
		// hThread = OpenThread(THREAD_ALL_ACCESS, TRUE, dwThreadId);
		// if (INVALID_HANDLE_VALUE == hThread) throw _raise_error(HRESULT_FROM_WIN32(GetLastError(), _T("OpenThread"));
		hThread = GetCurrentThread();  // pseudo handle, no close handle required
		if (!hThread) throw _raise_error(HRESULT_FROM_WIN32(GetLastError(), _T("GetCurrentThread"));

		BOOL br = OpenThreadToken(hThread, TOKEN_ADJUST_PRIVILEGES, TRUE, &hThdToken);
		if (!br || INVALID_HANDLE_VALUE == hThdToken) throw _raise_error(HRESULT_FROM_WIN32(GetLastError()), _T("OpenThreadToken"));
		*/

		// DWORD dwProcessId = GetCurrentProcessId();
		// if (!dwProcessId) throw _raise_error(HRESULT_FROM_WIN32(GetLastError()), _T("GetCurrentProcessId"));
		// hProcess = OpenProcess(PROCESS_ALL_ACCESS, TRUE, dwProcessId);
		// if (INVALID_HANDLE_VALUE == hProcess) throw _raise_error(HRESULT_FROM_WIN32(GetLastError()), _T("OpenProcess"));
		hProcess = GetCurrentProcess();  // pseudo handle, no close handle required
		if (!hProcess) throw _raise_error(HRESULT_FROM_WIN32(GetLastError()), _T("GetCurrentProcess"));

		BOOL br = OpenProcessToken(hProcess, TOKEN_QUERY | TOKEN_ADJUST_PRIVILEGES, &hPrcToken);
		if (!br || INVALID_HANDLE_VALUE == hPrcToken) throw _raise_error(HRESULT_FROM_WIN32(GetLastError()), _T("OpenProcessToken"));

		// note - SetSystemPowerState requires that the calling process/thread must have
		// the SE_SHUTDOWN_NAME privilege administratively authorized and enabled so we
		// will enable SeShutdownPrivilege, aka SE_SHUTDOWN_NAME
		TOKEN_PRIVILEGES oTokPrvOld;
		hr = EnablePrivilege(hPrcToken, SE_SHUTDOWN_NAME, TRUE, oTokPrvOld);
		if (FAILED(hr) || ERROR_NOT_ALL_ASSIGNED == hr) throw _raise_error(HRESULT_FROM_WIN32(GetLastError()), _T("_enablePrivilege"));
		// todo - check if ERROR_NOT_ALL_ASSIGNED will cause TRUE == FAILED(hr) and if so remove || from test

		br = SetSystemPowerState(TRUE,	  // system state - TRUE = suspend, FALSE = hibernate
								 FALSE);  // suspension option - TRUE = force, FALSE = request
		if (!br) throw _raise_error(HRESULT_FROM_WIN32(GetLastError()), _T("SetSystemPowerState"));

		hr = RestorePrivilege(hPrcToken, oTokPrvOld);
		if (FAILED(hr) || ERROR_NOT_ALL_ASSIGNED == hr) throw _raise_error(HRESULT_FROM_WIN32(GetLastError()), _T("_enablePrivilege"));
		// todo - check if ERROR_NOT_ALL_ASSIGNED will cause TRUE == FAILED(hr) and if so remove || from test

		// _TCHAR gzBuffer[MAX_PATH] = _T("\0");
		memset(gzBuffer, 0, sizeof(gzBuffer));
		_tcscpy(gzBuffer, _T("\nresuming graceful system suspend requested from a winsta"));
		switch (dwWinStaType) {
		case WINSTA_CONSOLE:
			_tcscat(gzBuffer, _T("0 gui console session\n"));
			break;
		case WINSTA_TERMINAL:
			_tcscat(gzBuffer, _T("0 gui terminal session\n"));
			break;
		case WINSTA_CONS_OR_TRMN:
			_tcscat(gzBuffer, _T("0 gui console or terminal session    \n"));
			break;
		case WINSTA_TELNET:
			_tcscat(gzBuffer, _T(" non-gui telnet session\n"));
			break;
		default:  // case WINSTA_UNKNOWN:
			_tcscat(gzBuffer, _T(" non-gui unknown session\n"));
			break;
		}

		// optionally don't bother with a resume message if requested was initiated from 
		// a winsta session type that may not still be around when the system resumes
		// if (WINSTA_TELNET && dwWinStaType && !(WINSTA_UNKNOWN && dwWinStaType)) {
			_tprintf(gzBuffer);
		// }

		throw _raise_error(S_OK, _T("The operation completed successfully"));
	}

	catch (_error oError) {  // expected exception handler
		#if defined(_DEBUG)
			_bstr_t sBuffer = oError.pgzApiDesc;
			HRESULT hr = oError.hr;
			_TCHAR gzBuffer[] = _T("0x01234567\0");
			sBuffer = _ltot(hr, gzBuffer, 16); sBuffer += _T(", "); sBuffer += oError.pgzApiDesc;
			OutputDebugString(sBuffer);
		#endif
		
		if (S_OK != oError.hr) {
			_TCHAR gzBuffer[MAX_PATH] = _T("\0");
			_stprintf(gzBuffer, _T("hr = %x, The operation did not complete successfully\n"), oError.hr);
			_tprintf(gzBuffer);
			
			#if defined(_DEBUG)
				if (WINSTA_CONSOLE == dwWinStaType) {
					hDsktopWnd = GetDesktopWindow();
					if (INVALID_HANDLE_VALUE == hDsktopWnd) throw _raise_error(HRESULT_FROM_WIN32(GetLastError()), _T("GetDesktopWindow"));
					// br = SetForegroundWindow(hDsktopWnd);
					// if (!br) throw _raise_error(HRESULT_FROM_WIN32(GetLastError()), _T("SetForgroundWindow"));
					MessageBox(hDsktopWnd, gzBuffer, _T("Suspend"), MB_SETFOREGROUND);
					// if (INVALID_HANDLE_VALUE != hDsktopWnd) BOOL br = CloseHandle(hDsktopWnd);  // not documented as needed for this handle type
					if (INVALID_HANDLE_VALUE != hDsktopWnd) hDsktopWnd = 0;  // CloseHandle() generates and exception
				}
			#endif
		}

		// if (INVALID_HANDLE_VALUE != hDsktopWnd) BOOL br = CloseHandle(hDsktopWnd);  // not documented as needed for this handle type
		if (INVALID_HANDLE_VALUE != hDsktopWnd) hDsktopWnd = 0;  // CloseHandle() generates an exception
		if (INVALID_HANDLE_VALUE != hThdToken) BOOL br = CloseHandle(hThdToken);
		if (INVALID_HANDLE_VALUE != hThread) BOOL br = CloseHandle(hThread);
		if (INVALID_HANDLE_VALUE != hPrcToken) BOOL br = CloseHandle(hPrcToken);
		if (INVALID_HANDLE_VALUE != hProcess) BOOL br = CloseHandle(hProcess);
		
		#if defined(_UNICODE)		
			// HGLOBAL hg = GlobalFree(agzArgV);  // api that msdn doc's suggest be replaced by new heap functions
			// if (hg) throw (S_FALSE);
			HRESULT hrr = HeapFreeEx(agzArgV);
			// if (FAILED(hfr)) throw (S_FALSE);
			agzArgV = NULL;  // so we don't accidentally figure it is still initialized later
		#endif

		return oError.hr;
	}

	catch (...) {  // [optional] unexpected exception handler
		_TCHAR gzBuffer[MAX_PATH] = _T("\0");
		_tcscpy(gzBuffer, _T("\nThe operation generated an unexpected exception\n"));
		_tprintf(gzBuffer);
		
		#if defined(_DEBUG)
			if (WINSTA_CONSOLE == dwWinStaType) {
				hDsktopWnd = GetDesktopWindow();
				if (INVALID_HANDLE_VALUE == hDsktopWnd) throw _raise_error(HRESULT_FROM_WIN32(GetLastError()), _T("GetDesktopWindow"));
				// br = SetForegroundWindow(hDsktopWnd);
				// if (!br) throw _raise_error(HRESULT_FROM_WIN32(GetLastError()), _T("SetForgroundWindow"));
				MessageBox(hDsktopWnd, gzBuffer, _T("Suspend"), MB_SETFOREGROUND);
				// if (INVALID_HANDLE_VALUE != hDsktopWnd) BOOL br = CloseHandle(hDsktopWnd);  // not documented as needed for this handle type
				if (INVALID_HANDLE_VALUE != hDsktopWnd) hDsktopWnd = 0;  // CloseHandle() generates and exception
			}
		#endif
		
		// if (INVALID_HANDLE_VALUE != hDsktopWnd) BOOL br = CloseHandle(hDsktopWnd);  // not documented as needed for this handle type
		if (INVALID_HANDLE_VALUE != hDsktopWnd) hDsktopWnd = 0;  // CloseHandle() generates an exception
		if (INVALID_HANDLE_VALUE != hThdToken) BOOL br = CloseHandle(hThdToken);
		if (INVALID_HANDLE_VALUE != hThread) BOOL br = CloseHandle(hThread);
		if (INVALID_HANDLE_VALUE != hPrcToken) BOOL br = CloseHandle(hPrcToken);
		if (INVALID_HANDLE_VALUE != hProcess) BOOL br = CloseHandle(hProcess);

		#if defined(_UNICODE)		
			// HGLOBAL hg = GlobalFree(agzArgV);  // api that msdn doc's suggest be replaced by new heap functions
			// if (hg) throw (S_FALSE);
			HRESULT hr = HeapFreeEx(agzArgV);
			// if (FAILED(hfr)) throw (S_FALSE);
			agzArgV = NULL;  // so we don't accidentally figure it is still initialized later
		#endif

		return E_FAIL;
	}
}
